# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

"""The imports in this modules are optional dependencies of pipedag or dependencies with multi-version support.

It is easiest in the pipedag codebase to optional dependencies them from this module as opposed to catching
ImportErrors across the codebase
"""
